/*
  mquake_recv.c - multicast earthquake early warning system
  written by Todd Hansen under NLANR/HPWREN
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <strings.h>
#include <time.h>
#include <sys/time.h>
#include "mquake.h"

#define RECEIVER_VER "1.4"

#define KEEPALIVE_TIMEOUT_DEFAULT default_keepalive*2+30

int debug, logfd;
char execf, execstr[80], coordf, coordstr[80], logf;
int threshold;
double last_thresh=0;

void version(void);
void usage(void);

void handle_keepalive(char *buf);
void handle_event(char *buf);
void handle_sensor(char *buf);

int init_connection(int mcast, int port, struct sockaddr_in *addr);

int main (int argc, char **argv)
{
  struct hostent *host_ent;
  struct sockaddr_in cli_addr;
  int sock_fd, lcv, lcv2;
  char mcast, buf[PKTSIZE+1];
  struct sockaddr_in addr;
  int port, keepalive_timeout;
  struct timeval timeout;
  fd_set fdset;
  unsigned long MTTL;
  char hostf, hoststr[80];
  int ch, arc;
  unsigned long lna;

  port = default_port;
  MTTL = default_MTTL;
  keepalive_timeout = KEEPALIVE_TIMEOUT_DEFAULT;
  debug=0;
  hostf=0;
  execf=0;
  logf=0;
  coordf=0;
  mcast=0;
  threshold=DEFAULT_THRESHOLD;
  while ((ch = getopt(argc, argv, "vhdt:l:c:k:p:m:e:")) != -1)
             switch (ch) {
	     case 'm':
	             strncpy(hoststr,optarg,80);
		     hostf=1;
	             break;
	     case 'e':
	             strncpy(execstr,optarg,80);
		     execf=1;
	             break;
	     case 'c':
	             strncpy(coordstr,optarg,80);
		     coordf=1;
	             break;
	     case 'l':
	             logfd=open(optarg,O_CREAT|O_APPEND|O_WRONLY|O_DIRECT);
		     if (logfd<0)
		       {
			 perror("open log file");
			 exit(-1);
		       }
		     logf=1;
	             break;
             case 'p':
                     port = atoi(optarg);
                     break;
             case 't':
                     threshold = atoi(optarg);
                     break;
             case 'k':
                     keepalive_timeout = atoi(optarg);
                     break;
             case 'v':
	             version();
                     break;
             case 'd':
	             debug=1;
                     break;
             case '?':
	     case 'h':
             default:
                     usage();
		     exit(-1);
             }
  argc -= optind;
  argv += optind;
  

  if (hostf)
    {
      if (-1 != (lna=inet_addr(hoststr)))
	memcpy(&addr.sin_addr, &lna, 
               min(sizeof(addr.sin_addr), sizeof(lna)));
      else 
	{
	  host_ent = gethostbyname(hoststr);
	  if (host_ent == NULL)
	    {
	      perror("gethostbyname");
	      fprintf(stderr,"couldn't find %s. exiting.\n",hoststr);
	      exit(-1);
	    }
	  memcpy(&addr.sin_addr, host_ent->h_addr, 
		 min(host_ent->h_length, sizeof(addr.sin_addr)));
	} 
      
      if ((ntohl(addr.sin_addr.s_addr) >= 0xe0000000) && (ntohl(addr.sin_addr.s_addr) <= 0xefffffff))
	{
	  if (debug)
	    printf("using multicast mode\n");
	  mcast=1;
	}
      else if (debug)
	printf("using UDP mode (unicast)\n");
    }
  else if (debug)
    printf("using UDP mode (unicast)\n");

  /*init socket*/
  sock_fd=init_connection(mcast,port,&addr);

  printf("MQuake ready.\n-------------\n");
  while (1) 
    {
      FD_ZERO(&fdset);
      FD_SET(sock_fd,&fdset);
      timeout.tv_usec=0;
      timeout.tv_sec=keepalive_timeout;

      arc=select(sock_fd+1,&fdset,NULL,NULL,&timeout);
      if (arc>0)
	{
	  lcv2=sizeof(cli_addr);
	  if ((lcv=recvfrom(sock_fd,buf,PKTSIZE,0,(struct sockaddr*)&cli_addr,&lcv2))<0)
	    {
	      perror("recvfrom");
	      exit(-1);
	    }
	  
	  buf[PKTSIZE]='\0';
	  for (lcv2=0;lcv2<PKTSIZE;lcv2++)
	    {
	      if (buf[lcv2]=='\n')
		{ buf[lcv2]='\0'; lcv2=PKTSIZE; }
	    }

	  if (debug)
	    {
	      printf("message from %s:\n-------\n%s\n-------\n",inet_ntoa(cli_addr.sin_addr),buf);
	      printf("received: %d bytes\n",lcv);
	    }
      
	  if (!strncmp(buf,"2315;",5))
	    {
	      switch (buf[5])
		{
		case 'K': 
		        handle_keepalive(buf+5);
			break;
		case 'E':
		        handle_event(buf+5);
			break;
		case 'S':
		        handle_sensor(buf+5);
			break;
		}
	    }
	  else if (debug)
	    printf("ignoring packet. magic number bad.\n");
	}
      else if (arc==0)
	{ /* argh, someone died! */
	  if (mcast)
	    {
	      if (debug)
		printf("%d. last 2 keepalives missing. initiating self-healing.\n",(int)time(NULL));
	     
	      close(sock_fd);
	      sleep(2);
	      sock_fd=init_connection(mcast,port,&addr);	     
	    }
	  else
	    {
	      fprintf(stderr,"%d. Last two keepalives not received!\n",(int)time(NULL));
	    }
	}
      else
	{
	  perror("select");
	  exit(-1);
	}
    }
  

  /* if we exit cleanly then remove group membership */
  /*  if (mcast)
    {
      mr.imr_multiaddr.s_addr = addr.sin_addr.s_addr;
      mr.imr_interface.s_addr = INADDR_ANY;
      if (setsockopt(sock_fd, IPPROTO_IP, IP_DROP_MEMBERSHIP,(char *)&mr, sizeof(mr)) < 0) 
	{
	  perror("IP_DROP_MEMBERSHIP");
	  exit(1);
	}
    }*/

  close(sock_fd);
  if (logf)
    close(logfd);

  return(0);
}

void usage(void)
{
  printf("Usage:\n\tmquake_recv [ -vhd ] [ -p port ] [ -k timeout ] [ -m Mcast_address ]\n\t\t    [ -c loc ] [ -l log ] [ -t threshold ] [ -e program ]\n\n\t\t-v\t\tVersion Information\n\t\t-d\t\tDebug Output\n\t\t-h\t\tShow usage information\n\t\t-p port\t\tchange the default port (default: %d)\n\t\t-k timeout\tchange the keepalive timeout (default: %d)\n\t\t-m src_host\tspecify a multicast src address\n\t\t-c loc\t\tlocation of you (format: 33.02658:-116.95945)\n\t\t-l log\t\tspecify a log file for logging events (append)\n\t\t-t threshold\tset the minimum threshold of a sensor event to\n\t\t\t\ttrigger a warning, 0 = all events trigger\n\t\t\t\t(default: %d)\n\t\t-e program\tmquake will call this program when an sensor is\n\t\t\t\ttriggered\n\nIf you are listening to a multicast address 'src_host' should be the address of\nthe multicast group. If you don't use this argument then mquake_recv will\noperate in unicast mode. 'program_name' will only be called when a sensor is\ntriggered and we receive notification. It will not be called when post-event\nanalysis is received. 'loc' is the latitude and longitude of your current\nlocation.  The values should be decimal coordinates in the form 'lat:long'.\n\n",default_port,KEEPALIVE_TIMEOUT_DEFAULT,DEFAULT_THRESHOLD);
}

void version(void)
{
  fprintf(stderr,"mquake_send v%s, protocol version v%s\n\n",RECEIVER_VER,PROTO_VER);
}


void handle_keepalive(char *buf)
{
  int lcv, server_time, server_utime;
  char *proto_ver=NULL;
  char *server_ver=NULL;
  char *server_time_ptr=NULL;
  char *server_utime_ptr=NULL;
  double timediff;
  struct timeval current_time;

  gettimeofday(&current_time,NULL);
  lcv=1;
  while(lcv<PKTSIZE && buf[lcv]!='\0')
    {
      if (buf[lcv]==';')
	{
	  if (server_time_ptr==NULL)
	    server_time_ptr=&(buf[lcv+1]);
	  else if (server_utime_ptr==NULL)
	    server_utime_ptr=&(buf[lcv+1]);
	  else if (proto_ver==NULL)
	    proto_ver=&(buf[lcv+1]);
	  else if (server_ver==NULL)
	    server_ver=&(buf[lcv+1]);
	  buf[lcv]='\0';
	}
      lcv++;
    }
  
  server_time=atoi(server_time_ptr);
  server_utime=atoi(server_utime_ptr);
  
  timediff=current_time.tv_sec-server_time;
  if (current_time.tv_usec >= server_utime)
    {
      timediff+=(current_time.tv_usec-server_utime)/1000000.0;
    }
  else
    {
      timediff+=(current_time.tv_usec+(1000000-server_utime))/1000000.0;
    }

  if (debug)
    {
      printf("%d:%d Keepalive received. timediff=%f. Protocol v%s, Server v%s\n",server_time,server_utime,timediff,proto_ver,server_ver);
      fflush(stdout);
    }

  if (strcmp(proto_ver,PROTO_VER))
    {
      fprintf(stderr,"Incompatible protocol version, client version %s, server version %s!\n",PROTO_VER,proto_ver);
      exit(-1);
    }
}

void handle_event(char *buf)
{
  double server_time, current_time_d;
  int lcv;
  char *amplitude_str=NULL;
  char *lat_str=NULL;
  char *long_str=NULL;
  char *elev_str=NULL;
  char *server_time_str=NULL;
  char *accuracy_str=NULL;
  struct timeval current_time;

  gettimeofday(&current_time,NULL);
  current_time_d=current_time.tv_sec+current_time.tv_usec/1000000.0;

  if (logf)
    write(logfd,buf,strlen(buf));

  lcv=1;
  while(lcv<PKTSIZE && buf[lcv]!='\0')
    {
      if (buf[lcv]==';')
	    {
	      if (amplitude_str==NULL)
		amplitude_str=&(buf[lcv+1]);
	      else if (lat_str==NULL)
		lat_str=&(buf[lcv+1]);
	      else if (long_str==NULL)
		long_str=&(buf[lcv+1]);
	      else if (elev_str==NULL)
		elev_str=&(buf[lcv+1]);
	      else if (server_time_str==NULL)
		server_time_str=&(buf[lcv+1]);
	      else if (accuracy_str==NULL)
		accuracy_str=&(buf[lcv+1]);
	      buf[lcv]='\0';
	    }
	  lcv++;
	}

  server_time=atof(server_time_str);
      
  if (debug)
    printf("%0.3f Event summary received. delay=%0.3f\n",server_time,current_time_d-server_time);
  else
    printf("%0.3f Event summary received.\n",server_time);

  printf("%0.3f Earthquake occured at: Lat=%s, Long=%s, Elev=%s: amplitude=%s, accuracy=%s.\n",server_time,lat_str,long_str,elev_str,amplitude_str,accuracy_str);
  fflush(stdout);

  buf[0]='\n';
  if (logf)
    write(logfd,buf,1);

}

void handle_sensor(char *buf)
{
  double server_time, current_time_d, amplitude;
  int lcv;
  char *amplitude_str=NULL;
  char *lat_str=NULL;
  char *long_str=NULL;
  char *elev_str=NULL;
  char *server_time_str=NULL;
  char *sensorid=NULL;
  struct timeval current_time;

  gettimeofday(&current_time,NULL);
  current_time_d=current_time.tv_sec+current_time.tv_usec/1000000.0;

  lcv=1;
  while(lcv<PKTSIZE && buf[lcv]!='\0')
    {
      if (buf[lcv]==';')
	{
	  if (amplitude_str==NULL)
	    amplitude_str=&(buf[lcv+1]);
	  else if (lat_str==NULL)
	    lat_str=&(buf[lcv+1]);
	  else if (long_str==NULL)
	    long_str=&(buf[lcv+1]);
	  else if (elev_str==NULL)
	    elev_str=&(buf[lcv+1]);
	  else if (server_time_str==NULL)
	    server_time_str=&(buf[lcv+1]);
	  else if (sensorid==NULL)
	    sensorid=&(buf[lcv+1]);
	  buf[lcv]='\0';
	}
      lcv++;
    }

  server_time=atof(server_time_str);
  amplitude=atof(amplitude_str);

  if ( amplitude >= threshold || (last_thresh-server_time < 2.5*60 && amplitude >= 0.75*threshold))
    {
      if (execf && last_thresh-server_time > 2.5*60)
	system(execstr);

      if (debug)
	printf("%0.3f Sensor has picked up a possible earthquake. delay=%0.3f\n",server_time,current_time_d-server_time);
      else
	printf("%0.3f Sensor has picked up a possible earthquake.\n",server_time);
    
      printf("%0.3f lat=%s,long=%s,elev=%s,amplitude=%s,sensorid=%s\n",server_time,lat_str,long_str,elev_str,amplitude_str,sensorid);
      
      if (amplitude >= threshold)
	last_thresh=amplitude;
    }
  else if (debug)
    {
      printf("sensor detection ignored (below threshold)\n");
    }
    
  fflush(stdout);

}


int init_connection(int mcast, int port, struct sockaddr_in *addr)
{
  struct sockaddr_in serv_addr;
  struct ip_mreq mr;
  int ch;
  int sock_fd;

  /*make socket*/
  sock_fd = socket(AF_INET, SOCK_DGRAM, 0);

  if (mcast)
    {
      /* avoid EADDRINUSE in bind, more than one client can listen on a host */
      ch=1;
#ifdef BSD
      if (setsockopt(sock_fd,SOL_SOCKET,SO_REUSEPORT,&ch,sizeof(ch))<0)
	{
	  perror("setsockopt: SO_REUSEPORT");
	  exit(-1);
	}
#endif
#ifndef BSD
      if (setsockopt(sock_fd,SOL_SOCKET,SO_REUSEADDR,&ch,sizeof(ch))<0)
	{
	  perror("setsockopt: SO_REUSEPORT");
	  exit(-1);
	}
#endif
    }

  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(port);
  
  if (bind(sock_fd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
    {
      perror("bind");
      exit(-1);
    }

  if (mcast)
    {
      mr.imr_multiaddr.s_addr = addr->sin_addr.s_addr;
      mr.imr_interface.s_addr = INADDR_ANY;
      if (setsockopt(sock_fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,(char *)&mr, sizeof(mr)) < 0) 
	{
	  perror("IP_ADD_MEMBERSHIP");
	  exit(1);
	}
    }

  return(sock_fd);
}



